@@ -1,10 +1,11 @@
 import re
+from contextlib import contextmanager
 
-from rope.base.utils.datastructures import OrderedSet
 from rope.base import ast, codeanalyze
 from rope.base.change import ChangeSet, ChangeContents
 from rope.base.exceptions import RefactoringError
 from rope.base.utils import pycompat
+from rope.base.utils.datastructures import OrderedSet
 from rope.refactor import (sourceutils, similarfinder,
                            patchedast, suites, usefunction)
 
@@ -35,6 +36,8 @@ from rope.refactor import (sourceutils, similarfinder,
 # classes.
 class _ExtractRefactoring(object):
 
+    kind_prefixes = {}
+
     def __init__(self, project, resource, start_offset, end_offset,
                  variable=False):
         self.project = project
@@ -52,33 +55,57 @@ class _ExtractRefactoring(object):
             offset -= 1
         return offset
 
-    def get_changes(self, extracted_name, similar=False, global_=False):
+    def get_changes(self, extracted_name, similar=False, global_=False, kind=None):
         """Get the changes this refactoring makes
 
         :parameters:
+            - `extracted_name`: target name, when starts with @ - set kind to
+            classmethod, $ - staticmethod
             - `similar`: if `True`, similar expressions/statements are also
               replaced.
             - `global_`: if `True`, the extracted method/variable will
               be global.
+            - `kind`: kind of target refactoring to (staticmethod, classmethod)
 
         """
+        extracted_name, kind = self._get_kind_from_name(extracted_name, kind)
+
         info = _ExtractInfo(
             self.project, self.resource, self.start_offset, self.end_offset,
-            extracted_name, variable=self.kind == 'variable',
+            extracted_name, variable=self._get_kind(kind) == 'variable',
             similar=similar, make_global=global_)
+        info.kind = self._get_kind(kind)
         new_contents = _ExtractPerformer(info).extract()
-        changes = ChangeSet('Extract %s <%s>' % (self.kind,
+        changes = ChangeSet('Extract %s <%s>' % (info.kind,
                                                  extracted_name))
         changes.add_change(ChangeContents(self.resource, new_contents))
         return changes
 
+    def _get_kind_from_name(self, extracted_name, kind):
+        for sign, selected_kind in self.kind_prefixes.items():
+            if extracted_name.startswith(sign):
+                self._validate_kind_prefix(kind, selected_kind)
+                return extracted_name[1:], selected_kind
+        return extracted_name, kind
+
+    @staticmethod
+    def _validate_kind_prefix(kind, selected_kind):
+        if kind and kind != selected_kind:
+            raise RefactoringError("Kind and shortcut in name mismatch")
+
+    @classmethod
+    def _get_kind(cls, kind):
+        raise NotImplementedError("You have to sublass {}".format(cls))
+
 
 class ExtractMethod(_ExtractRefactoring):
-
-    def __init__(self, *args, **kwds):
-        super(ExtractMethod, self).__init__(*args, **kwds)
-
     kind = 'method'
+    allowed_kinds = ("function", "method", "staticmethod", "classmethod")
+    kind_prefixes = {"@": "classmethod", "$": "staticmethod"}
+
+    @classmethod
+    def _get_kind(cls, kind):
+        return kind if kind in cls.allowed_kinds else cls.kind
 
 
 class ExtractVariable(_ExtractRefactoring):
@@ -90,6 +117,8 @@ class ExtractVariable(_ExtractRefactoring):
 
     kind = 'variable'
 
+    def _get_kind(cls, kind):
+        return cls.kind
 
 class _ExtractInfo(object):
     """Holds information about the extract to be performed"""
@@ -106,6 +135,7 @@ class _ExtractInfo(object):
         self.variable = variable
         self.similar = similar
         self._init_parts(start, end)
+        self.kind = None
         self._init_scope()
         self.make_global = make_global
 
@@ -182,16 +212,32 @@ class _ExtractInfo(object):
     def extracted(self):
         return self.source[self.region[0]:self.region[1]]
 
+    _cached_parsed_extraced = None
+
+    @property
+    def _parsed_extracted(self):
+        if self._cached_parsed_extraced is None:
+            self._cached_parsed_extraced = _parse_text(self.extracted)
+        return self._cached_parsed_extraced
+
     _returned = None
 
     @property
     def returned(self):
         """Does the extracted piece contain return statement"""
         if self._returned is None:
-            node = _parse_text(self.extracted)
-            self._returned = usefunction._returns_last(node)
+            self._returned = usefunction._returns_last(self._parsed_extracted)
         return self._returned
 
+    _returning_named_expr = None
+
+    @property
+    def returning_named_expr(self):
+        """Does the extracted piece contains named expression/:= operator)"""
+        if self._returning_named_expr is None:
+            self._returning_named_expr = usefunction._namedexpr_last(self._parsed_extracted)
+        return self._returning_named_expr
+
 
 class _ExtractCollector(object):
     """Collects information needed for performing the extract"""
@@ -368,7 +414,7 @@ class _ExceptionalConditionChecker(object):
         if end_scope != info.scope and end_scope.get_end() != end_line:
             raise RefactoringError('Bad region selected for extract method')
         try:
-            extracted = info.source[info.region[0]:info.region[1]]
+            extracted = info.extracted
             if info.one_line:
                 extracted = '(%s)' % extracted
             if _UnmatchedBreakOrContinueFinder.has_errors(extracted):
@@ -384,16 +430,24 @@ class _ExceptionalConditionChecker(object):
         if info.variable and not info.one_line:
             raise RefactoringError('Extract variable should not '
                                    'span multiple lines.')
+        if usefunction._named_expr_count(info._parsed_extracted) - usefunction._namedexpr_last(info._parsed_extracted):
+            raise RefactoringError('Extracted piece cannot '
+                                   'contain named expression (:= operator).')
 
     def multi_line_conditions(self, info):
         node = _parse_text(info.source[info.region[0]:info.region[1]])
         count = usefunction._return_count(node)
+        extracted = info.extracted
         if count > 1:
             raise RefactoringError('Extracted piece can have only one '
                                    'return statement.')
         if usefunction._yield_count(node):
             raise RefactoringError('Extracted piece cannot '
                                    'have yield statements.')
+        if not hasattr(ast, 'PyCF_ALLOW_TOP_LEVEL_AWAIT') and _AsyncStatementFinder.has_errors(extracted):
+            raise RefactoringError('Extracted piece can only have async/await '
+                                   'statements if Rope is running on Python '
+                                   '3.8 or higher')
         if count == 1 and not usefunction._returns_last(node):
             raise RefactoringError('Return should be the last statement.')
         if info.region != info.lines_region:
@@ -420,6 +474,26 @@ class _ExtractMethodParts(object):
     def __init__(self, info):
         self.info = info
         self.info_collector = self._create_info_collector()
+        self.info.kind = self._get_kind_by_scope()
+        self._check_constraints()
+
+    def _get_kind_by_scope(self):
+        if self._extacting_from_staticmethod():
+            return "staticmethod"
+        elif self._extracting_from_classmethod():
+            return "classmethod"
+        return self.info.kind
+
+    def _check_constraints(self):
+        if self._extracting_staticmethod() or self._extracting_classmethod():
+            if not self.info.method:
+                raise RefactoringError("Cannot extract to staticmethod/classmethod outside class")
+
+    def _extacting_from_staticmethod(self):
+        return self.info.method and _get_function_kind(self.info.scope) == "staticmethod"
+
+    def _extracting_from_classmethod(self):
+        return self.info.method and _get_function_kind(self.info.scope) == "classmethod"
 
     def get_definition(self):
         if self.info.global_:
@@ -473,10 +547,9 @@ class _ExtractMethodParts(object):
     def _get_function_definition(self):
         args = self._find_function_arguments()
         returns = self._find_function_returns()
+
         result = []
-        if self.info.method and not self.info.make_global and \
-           _get_function_kind(self.info.scope) != 'method':
-            result.append('@staticmethod\n')
+        self._append_decorators(result)
         result.append('def %s:\n' % self._get_function_signature(args))
         unindented_body = self._get_unindented_function_body(returns)
         indents = sourceutils.get_indent(self.info.project)
@@ -486,10 +559,22 @@ class _ExtractMethodParts(object):
 
         return definition + '\n'
 
+    def _append_decorators(self, result):
+        if self._extracting_staticmethod():
+            result.append('@staticmethod\n')
+        elif self._extracting_classmethod():
+            result.append('@classmethod\n')
+
+    def _extracting_classmethod(self):
+        return self.info.kind == "classmethod"
+
+    def _extracting_staticmethod(self):
+        return self.info.kind == "staticmethod"
+
     def _get_function_signature(self, args):
         args = list(args)
         prefix = ''
-        if self._extracting_method():
+        if self._extracting_method() or self._extracting_classmethod():
             self_name = self._get_self_name()
             if self_name is None:
                 raise RefactoringError('Extracting a method from a function '
@@ -501,44 +586,51 @@ class _ExtractMethodParts(object):
             '(%s)' % self._get_comma_form(args)
 
     def _extracting_method(self):
-        return self.info.method and not self.info.make_global and \
-            _get_function_kind(self.info.scope) == 'method'
+        return  not self._extracting_staticmethod() and \
+                (self.info.method and \
+                    not self.info.make_global and \
+                    _get_function_kind(self.info.scope) == 'method')
 
     def _get_self_name(self):
+        if self._extracting_classmethod():
+            return "cls"
+        return self._get_scope_self_name()
+
+    def _get_scope_self_name(self):
+        if self.info.scope.pyobject.get_kind() == "staticmethod":
+            return
         param_names = self.info.scope.pyobject.get_param_names()
         if param_names:
             return param_names[0]
 
     def _get_function_call(self, args):
+        return '{prefix}{name}({args})'.format(
+            prefix=self._get_function_call_prefix(args),
+            name=self.info.new_name,
+            args=self._get_comma_form(args))
+
+    def _get_function_call_prefix(self, args):
         prefix = ''
         if self.info.method and not self.info.make_global:
-            if _get_function_kind(self.info.scope) == 'method':
+            if self._extracting_staticmethod() or self._extracting_classmethod():
+                prefix = self.info.scope.parent.pyobject.get_name() + '.'
+            else:
                 self_name = self._get_self_name()
                 if self_name in args:
                     args.remove(self_name)
                 prefix = self_name + '.'
-            else:
-                prefix = self.info.scope.parent.pyobject.get_name() + '.'
-        return prefix + '%s(%s)' % (self.info.new_name,
-                                    self._get_comma_form(args))
+        return prefix
 
     def _get_comma_form(self, names):
-        result = ''
-        if names:
-            result += names[0]
-            for name in names[1:]:
-                result += ', ' + name
-        return result
+        return ', '.join(names)
 
     def _get_call(self):
-        if self.info.one_line:
-            args = self._find_function_arguments()
-            return self._get_function_call(args)
         args = self._find_function_arguments()
         returns = self._find_function_returns()
         call_prefix = ''
-        if returns:
-            call_prefix = self._get_comma_form(returns) + ' = '
+        if returns and (not self.info.one_line or self.info.returning_named_expr):
+            assignment_operator = ' := ' if self.info.one_line else ' = '
+            call_prefix = self._get_comma_form(returns) + assignment_operator
         if self.info.returned:
             call_prefix = 'return '
         return call_prefix + self._get_function_call(args)
@@ -558,14 +650,19 @@ class _ExtractMethodParts(object):
             return list(result)
         start = self.info.region[0]
         if start == self.info.lines_region[0]:
-            start = start + re.search('\S', self.info.extracted).start()
+            start = start + re.search('\\S', self.info.extracted).start()
         function_definition = self.info.source[start:self.info.region[1]]
         read = _VariableReadsAndWritesFinder.find_reads_for_one_liners(
             function_definition)
         return list(self.info_collector.prewritten.intersection(read))
 
     def _find_function_returns(self):
-        if self.info.one_line or self.info.returned:
+        if self.info.one_line:
+            written = self.info_collector.written | \
+                self.info_collector.maybe_written
+            return list(written & self.info_collector.postread)
+
+        if self.info.returned:
             return []
         written = self.info_collector.written | \
             self.info_collector.maybe_written
@@ -573,7 +670,10 @@ class _ExtractMethodParts(object):
 
     def _get_unindented_function_body(self, returns):
         if self.info.one_line:
-            return 'return ' + _join_lines(self.info.extracted)
+            if self.info.returning_named_expr:
+                return 'return ' + '(' + _join_lines(self.info.extracted) + ')'
+            else:
+                return 'return ' + _join_lines(self.info.extracted)
         extracted_body = self.info.extracted
         unindented_body = sourceutils.fix_indentation(extracted_body, 0)
         if returns:
@@ -615,6 +715,7 @@ class _FunctionInformationCollector(object):
         self.postwritten = OrderedSet()
         self.host_function = True
         self.conditional = False
+        self.loop_depth = 0
 
     def _read_variable(self, name, lineno):
         if self.start <= lineno <= self.end:
@@ -631,6 +732,8 @@ class _FunctionInformationCollector(object):
                 self.maybe_written.add(name)
             else:
                 self.written.add(name)
+            if self.loop_depth > 0 and name in self.read:
+                self.postread.add(name)
         if self.start > lineno:
             self.prewritten.add(name)
         if self.end < lineno:
@@ -651,6 +754,9 @@ class _FunctionInformationCollector(object):
             for name in visitor.read - visitor.written:
                 self._read_variable(name, node.lineno)
 
+    def _AsyncFunctionDef(self, node):
+        self._FunctionDef(node)
+
     def _Name(self, node):
         if isinstance(node.ctx, (ast.Store, ast.AugStore)):
             self._written_variable(node.id, node.lineno)
@@ -670,23 +776,15 @@ class _FunctionInformationCollector(object):
     def _ClassDef(self, node):
         self._written_variable(node.name, node.lineno)
 
-    def _handle_conditional_node(self, node):
-        self.conditional = True
-        try:
-            for child in ast.get_child_nodes(node):
-                ast.walk(child, self)
-        finally:
-            self.conditional = False
-
     def _If(self, node):
         self._handle_conditional_node(node)
 
     def _While(self, node):
-        self._handle_conditional_node(node)
+        with self._handle_loop_context(node):
+            self._handle_conditional_node(node)
 
     def _For(self, node):
-        self.conditional = True
-        try:
+        with self._handle_loop_context(node), self._handle_conditional_context(node):
             # iter has to be checked before the target variables
             ast.walk(node.iter, self)
             ast.walk(node.target, self)
@@ -695,9 +793,30 @@ class _FunctionInformationCollector(object):
                 ast.walk(child, self)
             for child in node.orelse:
                 ast.walk(child, self)
+
+    def _handle_conditional_node(self, node):
+        with self._handle_conditional_context(node):
+            for child in ast.get_child_nodes(node):
+                ast.walk(child, self)
+
+    @contextmanager
+    def _handle_conditional_context(self, node):
+        if self.start <= node.lineno <= self.end:
+            self.conditional = True
+        try:
+            yield
         finally:
             self.conditional = False
 
+    @contextmanager
+    def _handle_loop_context(self, node):
+        if node.lineno < self.start:
+            self.loop_depth += 1
+        try:
+            yield
+        finally:
+            self.loop_depth -= 1
+
 
 def _get_argnames(arguments):
     result = [pycompat.get_ast_arg_arg(node) for node in arguments.args
@@ -735,8 +854,6 @@ class _VariableReadsAndWritesFinder(object):
     def find_reads_and_writes(code):
         if code.strip() == '':
             return set(), set()
-        if isinstance(code, unicode):
-            code = code.encode('utf-8')
         node = _parse_text(code)
         visitor = _VariableReadsAndWritesFinder()
         ast.walk(node, visitor)
@@ -752,7 +869,18 @@ class _VariableReadsAndWritesFinder(object):
         return visitor.read
 
 
-class _UnmatchedBreakOrContinueFinder(object):
+class _BaseErrorFinder(object):
+    @classmethod
+    def has_errors(cls, code):
+        if code.strip() == '':
+            return False
+        node = _parse_text(code)
+        visitor = cls()
+        ast.walk(node, visitor)
+        return visitor.error
+
+
+class _UnmatchedBreakOrContinueFinder(_BaseErrorFinder):
 
     def __init__(self):
         self.error = False
@@ -792,14 +920,23 @@ class _UnmatchedBreakOrContinueFinder(object):
     def _ClassDef(self, node):
         pass
 
-    @staticmethod
-    def has_errors(code):
-        if code.strip() == '':
-            return False
-        node = _parse_text(code)
-        visitor = _UnmatchedBreakOrContinueFinder()
-        ast.walk(node, visitor)
-        return visitor.error
+
+class _AsyncStatementFinder(_BaseErrorFinder):
+
+    def __init__(self):
+        self.error = False
+
+    def _AsyncFor(self, node):
+        self.error = True
+
+    def _AsyncWith(self, node):
+        self.error = True
+
+    def _FunctionDef(self, node):
+        pass
+
+    def _ClassDef(self, node):
+        pass
 
 
 def _get_function_kind(scope):
@@ -808,7 +945,15 @@ def _get_function_kind(scope):
 
 def _parse_text(body):
     body = sourceutils.fix_indentation(body, 0)
-    node = ast.parse(body)
+    try:
+        node = ast.parse(body)
+    except SyntaxError:
+        # needed to parse expression containing := operator
+        try:
+            node = ast.parse('(' + body + ')')
+        except SyntaxError:
+            node = ast.parse('async def __rope_placeholder__():\n' + sourceutils.fix_indentation(body, 4))
+            node.body = node.body[0].body
     return node
 
 
